@@ -4,6 +4,7 @@ A function-oriented testing framework for Python 3.
 See README.md
 """
 import os
+import sys
 import time
 import inspect
 import types
@@ -13,8 +14,10 @@ import re
 import json
 import tempfile
 import shutil
-from contextlib import redirect_stdout, redirect_stderr
 import dataclasses
+import ctypes
+import tempfile
+from tempfile import NamedTemporaryFile
 from functools import wraps
 from contextlib import contextmanager
 from random import shuffle
@@ -32,13 +35,130 @@ def log(*args):
     if log_last_time is not None:
         delta = time.time() - log_last_time
     log_last_time = time.time()
-    log_fp.write(f"{delta:3.1f} " + "".join([str(i) + " " for i in args]) + "\n")
+    #log_fp.write(f"{delta:3.1f} " + "".join([str(i) + " " for i in args]) + "\n")
+    log_fp.write("".join([str(i) + " " for i in args]) + "\n")
     log_fp.flush()
 
 
 ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
 
 
+# Redirection is re-entrant and pauseable
+libc = ctypes.CDLL(None)
+redirect_depth = 0
+
+so_orig_fd = sys.stdout.fileno()  # The root level handle for stdout (typically == 1)
+so_root_save_fd = None  # Will be set for the root level stdout so that it can be used in pause
+so_c_fd = None  # The libc handle so that it can be flushed
+so_curr_tmpfile = None  # The top of the so stack tmpfile which is needed by pause/resume
+
+se_orig_fd = sys.stderr.fileno()  # The root level handle for stderr (typically == 2)
+se_root_save_fd = None  # Will be set for the root level stderr so that it can be used in pause
+se_c_fd = None  # The libc handle se that it can be flushed
+se_curr_tmpfile = None  # The top of the se stack tmpfile which is needed by pause/resume
+
+try:
+    # Linux
+    so_c_fd = ctypes.c_void_p.in_dll(libc, 'stdout')
+except ValueError:
+    # OSX
+    so_c_fd = ctypes.c_void_p.in_dll(libc, '__stdoutp')
+
+try:
+    # Linux
+    se_c_fd = ctypes.c_void_p.in_dll(libc, 'stderr')
+except ValueError:
+    # OSX
+    se_c_fd = ctypes.c_void_p.in_dll(libc, '__stderrp')
+
+
+def _redirect_stdout(to_fd):
+    libc.fflush(so_c_fd)
+    sys.stdout.close()
+    os.dup2(to_fd, so_orig_fd)
+    sys.stdout = io.TextIOWrapper(os.fdopen(so_orig_fd, "wb"))
+
+
+def _redirect_stderr(to_fd):
+    libc.fflush(se_c_fd)
+    sys.stderr.close()
+    os.dup2(to_fd, se_orig_fd)
+    sys.stderr = io.TextIOWrapper(os.fdopen(se_orig_fd, "wb"))
+
+
+@contextmanager
+def stdio_capture(should_capture):
+    """
+    Capture stdout in a re-entrant manner. See pause_stdio_capture().
+
+    If should_capture is False it simply returns (stdout, stderr)
+    which simplifies conditional "with" clauses. Ie:
+
+        with stdio_capture(should_capture) as (so, se):
+            important_stuff(so, se)
+
+    as opposed to:
+
+        if should_capture:
+            with stdio_capture(should_capture) as (so, se):
+                important_stuff(so, se)
+        else:
+            # repeating the above
+            important_stuff(sys.stdout, sys.stderr)
+
+    """
+
+    if not should_capture:
+        yield sys.stdout, sys.stderr
+    else:
+        global redirect_depth
+        global so_root_save_fd, so_curr_tmpfile
+        global se_root_save_fd, se_curr_tmpfile
+
+        so_save_fd = os.dup(so_orig_fd)
+        se_save_fd = os.dup(se_orig_fd)
+        if redirect_depth == 0:
+            so_root_save_fd = so_save_fd
+            se_root_save_fd = se_save_fd
+
+        so_tmpfile = NamedTemporaryFile(mode="w+")
+        se_tmpfile = NamedTemporaryFile(mode="w+")
+
+        so_prev_tmpfile = so_curr_tmpfile
+        se_prev_tmpfile = se_curr_tmpfile
+
+        so_curr_tmpfile = so_tmpfile
+        se_curr_tmpfile = se_tmpfile
+
+        redirect_depth += 1
+        try:
+            _redirect_stdout(so_tmpfile.fileno())
+            _redirect_stderr(se_tmpfile.fileno())
+            yield (so_tmpfile, se_tmpfile)
+            _redirect_stderr(se_save_fd)
+            _redirect_stdout(so_save_fd)
+        finally:
+            redirect_depth -= 1
+            so_tmpfile.close()
+            se_tmpfile.close()
+            so_curr_tmpfile = so_prev_tmpfile
+            se_curr_tmpfile = se_prev_tmpfile
+            os.close(so_save_fd)
+            os.close(se_save_fd)
+
+
+@contextmanager
+def pause_stdio_capture():
+    if redirect_depth > 0:
+        _redirect_stdout(so_root_save_fd)
+        _redirect_stderr(se_root_save_fd)
+        yield
+        _redirect_stdout(so_curr_tmpfile.fileno())
+        _redirect_stderr(se_curr_tmpfile.fileno())
+    else:
+        yield
+
+
 def strip_ansi(line):
     return ansi_escape.sub("", line)
 
@@ -281,7 +401,7 @@ class zest:
     _tmp_root = None
 
     @staticmethod
-    def reset(disable_shuffle=False, bypass_skip=None, common_tmp=None, tmp_root=None):
+    def reset(disable_shuffle=False, bypass_skip=None, common_tmp=None, tmp_root=None, capture=None):
         zest._call_log = []
         zest._call_stack = []
         zest._call_errors = []
@@ -291,7 +411,7 @@ class zest:
         zest._test_stop_callback = None
         zest._mock_stack = []
         zest._allow_to_run = None
-        zest._capture = False
+        zest._capture = capture
         zest._disable_shuffle = disable_shuffle
         zest._bypass_skip = [] if bypass_skip is None else bypass_skip.split(":")
         zest._common_tmp = common_tmp
@@ -435,12 +555,14 @@ class zest:
 
     @staticmethod
     @contextmanager
-    def mock(symbol, returns=None):
+    def mock(symbol, returns=None, hook=None):
         old, klass, new = None, None, None
         try:
             old, klass, new = zest._setup_mock(symbol)
             if returns is not None:
                 new.returns(returns)
+            if hook is not None:
+                new.hook(hook)
             yield new
         finally:
             if klass and old:
@@ -614,150 +736,160 @@ class zest:
                     "A _begin function was declared. Maybe you meant _before?"
                 )
 
-            def _run(name, func, so, se):
-                # params_list = [tuple()]
-                # if hasattr(func, "params_list"):
-                #     params_list = getattr(func, "params_list")
-
-                zest._call_stack += [name]
-                zest._current_error = None
-
-                try:
-                    full_name = ".".join(zest._call_stack)
-                    if (
-                        zest._allow_to_run is not None
-                        and full_name not in zest._allow_to_run
-                        and zest._allow_to_run != "__all__"
-                    ):
-                        zest._call_stack.pop()
-                        return
-
-                except Exception as e:
-                    log(f"EXCEPTION during allow to check run. NAME {name} e {e}")
-                    zest._call_stack.pop()
-                    return
-
-                pre_cwd = os.getcwd()
-                remove_tmp_dir = None
-                try:
-                    if zest._common_tmp is not None:
-                        cwd = zest._common_tmp
-                    else:
-                        # Create a tmp folder per test
-                        tmp_root = zest._tmp_root or "/tmp"
-                        cwd = tempfile.mkdtemp(dir=tmp_root)
-                        remove_tmp_dir = cwd
-
-                    # Set each test into the correct tmp folder
-                    os.chdir(cwd)
-
-                    # for params in params_list:
-                    _before = callers_special_local_funcs.get("_before")
-                    if _before:
-                        try:
-                            _before()
-                        except Exception as e:
-                            zest._call_errors += [(e, zest._call_stack.copy())]
-                            s = (
-                                f"There was an exception while running '_before()' in test '{name}'. "
-                                f"This may mean that the sub-tests are not enumerated and therefore can not be run."
-                            )
-                            zest._call_warnings += [s]
-
-                    try:
-                        zest._call_tree += [full_name]
-                        zest._call_log += [full_name]
-
-                        if zest._test_start_callback:
-                            zest._test_start_callback(
-                                ZestResult(
-                                    zest._call_stack,
-                                    full_name,
-                                    zest._call_stack[-1],
-                                    None,
-                                    None,
-                                    None,
-                                    None,
-                                    None,
-                                    None,
-                                    func.__code__.co_filename,
-                                    os.getpid(),
-                                    True,
-                                )
-                            )
-
-                        error = None
-                        error_formatted = None
-                        skip_reason = None
-                        start_time = time.time()
-                        try:
-                            zest._mock_stack += [[]]
-
-                            try:
-                                func()
-                            except SkipException as e:
-                                skip_reason = e.reason
-                            zest._clear_stack_mocks()
-                            zest._mock_stack.pop()
-                        except Exception as e:
-                            error = e
-                            error_formatted = traceback.format_exception(
-                                etype=type(error), value=error, tb=error.__traceback__
-                            )
-                            zest._call_errors += [
-                                (e, error_formatted, zest._call_stack.copy())
-                            ]
-                            zest._current_error = e
-                        finally:
-                            stop_time = time.time()
-                            if zest._test_stop_callback:
-                                zest_result = ZestResult(
-                                    zest._call_stack,
-                                    ".".join(zest._call_stack),
-                                    zest._call_stack[-1],
-                                    error,
-                                    error_formatted,
-                                    stop_time - start_time,
-                                    skip_reason,
-                                    so.getvalue() if so is not None else None,
-                                    se.getvalue() if se is not None else None,
-                                    func.__code__.co_filename,
-                                    os.getpid(),
-                                    False,
-                                )
-                                zest._test_stop_callback(zest_result)
-
-                        _after = callers_special_local_funcs.get("_after")
-                        if _after:
-                            _after()
-                    except Exception as e:
-                        log(f"ZEST EXCEPTION 1 {e}")
-                    finally:
-                        zest._call_stack.pop()
-                finally:
-                    # Clean up tmp folders if needed
-                    if remove_tmp_dir:
-                        try:
-                            shutil.rmtree(remove_tmp_dir)
-                        except OSError as exc:
-                            if exc.errno != errno.ENOENT:  # ENOENT - no such file or directory
-                                raise  # re-raise exception
-                    os.chdir(pre_cwd)
-
             for name, func in funcs_to_call:
                 if len(zest._mock_stack) > 0:
                     for mock_tuple in zest._mock_stack[-1]:
                         if mock_tuple[4]:  # if reset_before_each is set
                             mock_tuple[3].reset()  # Tell the mock to reset
 
-                if zest._capture:
-                    so = io.StringIO()
-                    se = io.StringIO()
-                    with redirect_stdout(so):
-                        with redirect_stderr(se):
-                            _run(name, func, so, se)
-                else:
-                    _run(name, func, None, None)
+                with stdio_capture(zest._capture) as (so, se):
+                    zest._call_stack += [name]
+                    zest._current_error = None
+
+                    try:
+                        full_name = ".".join(zest._call_stack)
+                        if (
+                            zest._allow_to_run is not None
+                            and full_name not in zest._allow_to_run
+                            and zest._allow_to_run != "__all__"
+                        ):
+                            zest._call_stack.pop()
+                            continue
+
+                    except Exception as e:
+                        log(f"EXCEPTION during allow to check run. NAME {name} e {e}")
+                        zest._call_stack.pop()
+                        continue
+
+                    pre_cwd = os.getcwd()
+                    remove_tmp_dir = None
+                    try:
+                        if zest._common_tmp is not None:
+                            cwd = zest._common_tmp
+                        else:
+                            # Create a tmp folder per test
+                            tmp_root = zest._tmp_root or "/tmp"
+                            cwd = tempfile.mkdtemp(dir=tmp_root)
+                            remove_tmp_dir = cwd
+
+                        # Set each test into the correct tmp folder
+                        os.chdir(cwd)
+
+                        # for params in params_list:
+                        _before = callers_special_local_funcs.get("_before")
+                        if _before:
+                            try:
+                                _before()
+                            except Exception as e:
+                                zest._call_errors += [(e, zest._call_stack.copy())]
+                                s = (
+                                    f"There was an exception while running '_before()' in test '{name}'. "
+                                    f"This may mean that the sub-tests are not enumerated and therefore can not be run."
+                                )
+                                zest._call_warnings += [s]
+
+                        try:
+                            zest._call_tree += [full_name]
+                            zest._call_log += [full_name]
+
+                            if zest._test_start_callback:
+                                with pause_stdio_capture():
+                                    zest._test_start_callback(
+                                        ZestResult(
+                                            zest._call_stack,
+                                            full_name,
+                                            zest._call_stack[-1],
+                                            None,
+                                            None,
+                                            None,
+                                            None,
+                                            None,
+                                            None,
+                                            func.__code__.co_filename,
+                                            os.getpid(),
+                                            True,
+                                        )
+                                    )
+
+                            error = None
+                            error_formatted = None
+                            skip_reason = None
+                            start_time = time.time()
+                            try:
+                                zest._mock_stack += [[]]
+
+                                try:
+                                    func()
+                                except SkipException as e:
+                                    skip_reason = e.reason
+                                zest._clear_stack_mocks()
+                                zest._mock_stack.pop()
+                            except Exception as e:
+                                error = e
+                                error_formatted = traceback.format_exception(
+                                    etype=type(error), value=error, tb=error.__traceback__
+                                )
+                                zest._call_errors += [
+                                    (e, error_formatted, zest._call_stack.copy())
+                                ]
+                                zest._current_error = e
+                            finally:
+                                stop_time = time.time()
+
+                                sys.stdout.flush()
+                                sys.stderr.flush()
+                                so.flush()
+                                se.flush()
+                                captured_so = None
+                                try:
+                                    so.seek(0, io.SEEK_SET)
+                                    captured_so = so.read()
+                                except io.UnsupportedOperation:
+                                    # This happens if so is actually sys.stdout
+                                    pass
+
+                                captured_se = None
+                                try:
+                                    se.seek(0, io.SEEK_SET)
+                                    captured_se = se.read()
+                                except io.UnsupportedOperation:
+                                    # This happens if se is actually sys.stderr
+                                    pass
+
+                                if zest._test_stop_callback:
+                                    zest_result = ZestResult(
+                                        zest._call_stack,
+                                        ".".join(zest._call_stack),
+                                        zest._call_stack[-1],
+                                        error,
+                                        error_formatted,
+                                        stop_time - start_time,
+                                        skip_reason,
+                                        captured_so if captured_so is not None else None,
+                                        captured_se if captured_se is not None else None,
+                                        func.__code__.co_filename,
+                                        os.getpid(),
+                                        False,
+                                    )
+                                    with pause_stdio_capture():
+                                        zest._test_stop_callback(zest_result)
+
+                            _after = callers_special_local_funcs.get("_after")
+                            if _after:
+                                _after()
+                        except Exception as e:
+                            log(f"ZEST EXCEPTION 1 {e}")
+                        finally:
+                            zest._call_stack.pop()
+                    finally:
+                        # Clean up tmp folders if needed
+                        if remove_tmp_dir:
+                            try:
+                                shutil.rmtree(remove_tmp_dir)
+                            except OSError as exc:
+                                if exc.errno != errno.ENOENT:  # ENOENT - no such file or directory
+                                    raise  # re-raise exception
+                        os.chdir(pre_cwd)
 
         finally:
             if prev_test_start_callback is not None:
