@@ -2,7 +2,9 @@ from zest import colors
 import sys
 import os
 import re
+import io
 import traceback
+from zest.zest import log
 
 
 def s(*strs):
@@ -10,6 +12,7 @@ def s(*strs):
         if str_ is not None:
             sys.stdout.write(str_)
     sys.stdout.write(colors.reset)
+    sys.stdout.flush()
 
 
 _tb_pat = re.compile(r"^.*File \"([^\"]+)\", line (\d+), in (.*)")
